from sepbox import *
from math import fabs


def ininterval(x,y):  # contractor associated to the contraint 'x is inside the interval y'
    ra=x&Interval(y.lb,y.lb)
    rb=x&Interval(y.ub,y.ub)
    r=ra|rb
    if x.lb in y: r.ca=1
    else: r.ca=0
    if x.ub in y: r.cb=1
    else: r.cb=0
    return r
    
def cadd_fwd(x,y):
    return x+y	

def cadd_bwd(z,x,y):
    x0=x.copy()
    y0=y.copy()
    x=x&z-y
    y=y&z-x
    if x0.lb<x.lb: x.ca=z.ca
    if x0.ub>x.ub: x.cb=z.cb
    if y0.lb<y.lb: y.ca=z.ca
    if y0.ub>y.ub: y.cb=z.cb
    if z.is_empty():
        x.ca,x.cb=z.ca,z.cb
        y.ca,y.cb=z.ca,z.cb
    return x,y	

def csqr_fwd(x):
    y=sqr(x)
    return y

def csqr_bwd(y,x):
    if y.is_empty():   return y
    x0=x.copy()
    x1=x&Interval(-oo,0) 
    x2=x&Interval(0,oo) 	
    x1=x1&(-1*sqrt(y))
    x2=x2&(sqrt(y))
    x=x1|x2
    if (x0.lb<x.lb):  x.ca=y.ca if ((x0.lb)**2<y.lb) else y.cb
    if (x0.ub>x.ub):  x.cb=y.ca if ((x0.ub)**2<y.lb) else y.cb
    return x



def C3(x,y):
    x2=csqr_fwd(x)
    y2=csqr_fwd(y)
    z=cadd_fwd(x2,y2)
    z=ininterval(z,Interval(1,4))
    x2,y2=cadd_bwd(z,x2,y2)
    x=csqr_bwd(x2,x)
    y=csqr_bwd(y2,y)
    return x,y


def C2(x,y):
    z=cadd_fwd(x,y)
    z=ininterval(z,Interval(-1,2))
    x,y=cadd_bwd(z,x,y)
    return x,y

def Cand2(C1,C2,x,y):
    x1,y1=C1(x,y)
    x2,y2=C2(x,y)
    rx=x.copy()
    ry=y.copy()
    if x1.is_empty() | y1.is_empty():
        if x1.ca==0: return Interval(1,0,x1.ca,x1.ca),Interval(1,0,x1.ca,x1.ca)
        if x1.ca==1: return x2,y2
    if x2.is_empty() | y2.is_empty():
        if x2.ca==0: return Interval(1,0,x2.ca,x2.ca),Interval(1,0,x2.ca,x2.ca)
        if x2.ca==1: return x1,y1
    if (x1.ca==0): rx.lb=max(x1.lb,rx.lb); rx.ca=0
    if (x2.ca==0): rx.lb=max(x2.lb,rx.lb); rx.ca=0
    if (x1.cb==0): rx.ub=min(x1.ub,rx.ub); rx.cb=0
    if (x2.cb==0): rx.ub=min(x2.ub,rx.ub); rx.cb=0
    if (y1.ca==0): ry.lb=max(y1.lb,ry.lb); ry.ca=0
    if (y2.ca==0): ry.lb=max(y2.lb,ry.lb); ry.ca=0
    if (y1.cb==0): ry.ub=min(y1.ub,ry.ub); ry.cb=0
    if (y2.cb==0): ry.ub=min(y2.ub,ry.ub); ry.cb=0
    if (x1.ca==1) and (x2.ca==1): rx.lb=min(x1.lb,x2.lb); rx.ca=1
    if (x1.cb==1) and (x2.cb==1): rx.ub=max(x1.ub,x2.ub); rx.cb=1
    if (y1.ca==1) and (y2.ca==1): ry.lb=min(y1.lb,y2.lb); ry.ca=1
    if (y1.cb==1) and (y2.cb==1): ry.ub=max(y1.ub,y2.ub); ry.cb=1
    return rx,ry
    

def Cand(C1,C2):
    C = lambda x,y : Cand2(C1,C2,x,y)
    return C

def C1(x,y):
    x=ininterval(x,Interval(1.1,2.2))
    y=ininterval(y,Interval(1.1,2.2))
    return x,y

def sivia(X):
    X0=X.copy()
    C=Cand(C2,C3)
    X.x,X.y = C(X.x,X.y)
    drawDiffBox(X0,X,True)
    if (X.width()<0.1):
        if (X.x.ca+X.x.cb==1) or (X.y.ca+X.y.cb==1):
            vibes.drawBox(X.x.lb, X.x.ub,X.y.lb, X.y.ub, '[orange]')
        else: vibes.drawBox(X.x.lb, X.x.ub,X.y.lb, X.y.ub, '[yellow]')
        return
    sivia(X.left())
    sivia(X.right())

if __name__ == '__main__':
    i=0
    vibes.beginDrawing()
    vibes.newFigure('Result')
    vibes.setFigureProperties({'x': 0, 'y': 0, 'width': 1600, 'height': 1000})
    X=Box(Interval(-3,3),Interval(-3,3))
    sivia(X)
    print('---- FINI -----')

    vibes.endDrawing()
